# -*- coding: utf-8 -*-
"""
Created on Tue Mar 15 17:34:25 2016

@author: Alexander
"""

import time
import monitorProgress
import numpy as np
from sklearn.metrics import zero_one_loss, roc_auc_score
from copy import copy

'''
#==============================================================================
# ARB algorithm
#==============================================================================
'''
class ARBClassifier(object):
    def __init__(self, base_estimator = None, n_estimators = 50, 
                 learning_rate = 1.0, weight_update = None, procedure = False):
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.weight_update = weight_update
        self.procedure = procedure
        
    # fit on X_train and predict X_test
    def fit_predict(self, X_train, y_train, X_test, y_test, 
                    sample_weight = None, gamma = 2, auc = False):
        M = len(y_train)
        decision_function = np.zeros(M)
        decision_function_test = np.zeros(len(y_test))
        
        if sample_weight == None:
            # uniform initial weights
            distribution = np.ones(M)*1.0/M
        else:
            distribution = sample_weight
            
        train_error = []
        test_error = []
            
        for iteration in range(self.n_estimators):
            if self.procedure:
                time.sleep(0.1)
                monitorProgress.update_progress((iteration+1)/(self.n_estimators*1.0))
        
            # fit decision stump using training data
            dt_stump = copy(self.base_estimator)
            dt_stump.fit(X_train, y_train, sample_weight = distribution)
    
            # estimate class probabilities using fitted weak learner
            prob = dt_stump.predict_proba(X_train)
            prob_test = dt_stump.predict_proba(X_test)
            
            # rescale the probability
            prob = 0.9999999*prob + 0.00000005
            prob_test = 0.9999999*prob_test + 0.00000005
    
            # calculate log-probabilities
            plog = np.log(prob)
            plog_test = np.log(prob_test)
    
            # calculate weak hypothesis h_t (ARB)
            weak_hypo = self.learning_rate*(plog[:,1] - plog[:,0])
            weak_hypo_test = self.learning_rate*(plog_test[:,1] - plog_test[:,0]) 
            
            # update decision function F_t
            decision_function += weak_hypo
            decision_function_test += weak_hypo_test
            #decision_function[np.negative(np.isinf(decision_function))] += weak_hypo[np.negative(np.isinf(decision_function))]
            #decision_function_test[np.negative(np.isinf(decision_function_test))] += weak_hypo_test[np.negative(np.isinf(decision_function_test))]
        
            distribution = self.weight_update(y_train, decision_function, gamma)
            
            prediction = np.sign(decision_function)
            prediction_test = np.sign(decision_function_test)
            
            if auc:
                prediction_proba = 1.0/(1 + np.exp(-2*decision_function))
                prediction_proba_test = 1.0/(1 + np.exp(-2*decision_function_test))
                train_error.append(roc_auc_score(y_train, prediction_proba))
                test_error.append(roc_auc_score(y_test, prediction_proba_test))
            else:
                train_error.append(zero_one_loss(prediction, y_train))
                test_error.append(zero_one_loss(prediction_test, y_test))
        
        if self.procedure:
            return train_error, test_error
        else:
            return train_error[-1], test_error[-1]